package rayCifePlus;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Random;
import java.util.Stack;

import objetosAuxilidares.*;
import util.*;

public class RayCifePath {

	private Arquivo arq;

	private Scene scene;

	private Screen screen; // Estrutura da tela
	private int numPaths;
	private double toneMapping; // O valor a ser utilizado no Tone Mapping
	Random random;
	int seed;

	private Stack<Integer> elemStack;

	public RayCifePath(String inputFile) {
		this.arq = new Arquivo(inputFile);
		this.scene = new Scene();
		this.screen = new Screen();
		this.numPaths = 5;
		this.toneMapping = 0;
		this.random = null;
		this.elemStack = new Stack<Integer>();
	}

	// Ler sdl
	private void readArqPrinc() {

		while (!arq.isEndOfFile()) {
			String str = arq.readString();
			if (str.charAt(0) == '#') {
				// Pular linha
				if (!arq.isEndOfLine()) {
					continue;
				}
			} else {
				if (str.equals("output")) {
					try {
						arq.criarSaida(arq.readString());
					} catch (Exception e) {
						// Do something here?
					}
				} else if (str.equalsIgnoreCase("eye")) {
					scene.setEye(new Point(arq.readDouble(), arq.readDouble(),
							arq.readDouble()));
				} else if (str.equalsIgnoreCase("ortho")) {
					screen.setZero(new Coord2d(arq.readInt(), arq.readInt()));
					screen.setOne(new Coord2d(arq.readInt(), arq.readInt()));
				} else if (str.equalsIgnoreCase("size")) {
					screen.initArray(arq.readInt(), arq.readInt());
				} else if (str.equalsIgnoreCase("background")) {
					scene.setBack(new Color(arq.readDouble(), arq.readDouble(),
							arq.readDouble()));
				} else if (str.equalsIgnoreCase("ambient")) {
					scene.setAmbient(arq.readDouble());
				} else if (str.equalsIgnoreCase("light")) {
					str = arq.readString();
					Triangulo geom = readObj(str);
					Color c = new Color(arq.readDouble(), arq.readDouble(), arq
							.readDouble());
					Light lTemp = new Light(geom, c, arq.readDouble());
					scene.addLight(lTemp);
				} else if (str.equalsIgnoreCase("npaths")) {
					this.numPaths = arq.readInt();
				} else if (str.equalsIgnoreCase("tonemapping")) {
					this.toneMapping = arq.readDouble();
				} else if (str.equalsIgnoreCase("seed")) {
					int aux = arq.readInt();
					this.random = new Random(aux);
					this.seed = aux;
				} else if (str.equalsIgnoreCase("objectquadric")) {
					CoeficendesDasQuadricas qc = new CoeficendesDasQuadricas(arq.readDouble(), arq
							.readDouble(), arq.readDouble(), arq.readDouble(),
							arq.readDouble(), arq.readDouble(), arq
									.readDouble(), arq.readDouble(), arq
									.readDouble(), arq.readDouble());
					Material mTemp = new Material(arq.readDouble(), arq.readDouble(), arq
							.readDouble(), arq.readDouble(), arq.readDouble(),
							arq.readDouble(), arq.readDouble(), arq.readInt(),
							arq.readDouble());
					Elemento elem = new Quadricas(qc, mTemp);
					List<Elemento> quad = new ArrayList<Elemento>();
					quad.add(elem);
					scene.addObject(quad);
				} else if (str.equalsIgnoreCase("object")) {
					str = arq.readString();
					System.out.println(str);
					Triangulo geom = readObj(str);
					Material mTemp = new Material(arq.readDouble(), arq.readDouble(), arq
							.readDouble(), arq.readDouble(), arq.readDouble(),
							arq.readDouble(), arq.readDouble(), arq.readInt(),
							arq.readDouble());
					List<Triangulo> objects = geom.getObjects();
					List<Elemento> elms = new ArrayList<Elemento>();
					for (Triangulo o : objects) {
						Elemento elem = new ElementoTriangulo(o, mTemp);
						elms.add(elem);
					}
					scene.addObject(elms);

				}
			}

		}
	} // </readSdl>

	// Leitura do arquivo obj
	private static Triangulo readObj(String filename) {
		List<Point> ps = new ArrayList<Point>();
		Triangulo result = new Triangulo(ps, new ArrayList<Face>());
		Arquivo arq = new Arquivo(filename);
		char c;
		Point p;
		Face f;
		while (!arq.isEndOfFile()) {
			c = arq.readChar();
			switch (c) {
			case '#':
				arq.readLine();
				break;
			case 'v':
				p = new Point(arq.readDouble(), arq.readDouble(), arq
						.readDouble());
				ps.add(p);
				break;
			case 'f':
				f = new Face(ps.get(arq.readInt() - 1), ps
						.get(arq.readInt() - 1), ps.get(arq.readInt() - 1));
				result.addFace(f);
				break;
			default:
				break;
			}
		}
		arq.close();
		return result;
	}

	public void executePathTracing() {
		this.readArqPrinc();
		this.loadMatrix();
		this.desenharImagem();
	}

	// Matriz de cores
	public void loadMatrix() {

		final Point eye = scene.getEye();
		final Coord2d zero = screen.getZero();
		final Coord2d one = screen.getOne();

		final int worldW = one.getX() - zero.getX();
		final int worldH = one.getY() - zero.getY();

		final int screenW = screen.getWidth();
		final int screenH = screen.getHeight();

		final double incW = ((double) worldW) / ((double) screenW);
		final double incH = ((double) worldH) / ((double) screenH);

		final double zerox = zero.getX() + incW / 2;
		final double zeroy = one.getY() - incH / 2;

		for (int i = 0; i < screenH; i++) {
			System.out.println("Rendering line " + i);
			for (int j = 0; j < screenW; j++) {				
				Vetor n = new Vetor(eye, new Point(zerox + j * incW, zeroy - i
						* incH, 0));
				n.normalize();
				Ray ray = new Ray(eye, eye.getAddition(n)); //??
				
				Color finalColor = new Color(0.0,0.0,0.0);
				boolean lightIntercepted = false;

				
				int pathNumber = 0;
				lightIntercepted = false; //
				for (pathNumber = 0; pathNumber < this.numPaths; pathNumber++) {
					
					elemStack.push(-1);
					
						Color temp = pathTrace(ray, 0);
						finalColor.somar(temp);
						if(temp.getR()==1.0 && temp.getG()==1.0 && temp.getB()==1.0){
							lightIntercepted = true; //
							break;
						
					}
					elemStack.clear();
				}
				finalColor.dividir(pathNumber);
					
				
				if (pathNumber == 1);
				if(!lightIntercepted){
					finalColor.dividir(finalColor.getSoma(this.toneMapping));
				}
				

				screen.setColorAt(finalColor, i, j);

			}
		}
	} //  </LoadMatrix>

	private Color pathTrace(final Ray ray, int secondaryRay){
		
		//System.out.println(secondaryRay);
		int elemTemp = 0;

		final Iterator<Elemento> elements = scene.getObjIt();
		Elemento closestElement = null;
		Point intersectionPoint = null;
		double minSquaredDist = Double.MAX_VALUE;
		
		// Pega o objeto mais proximo
		while (elements.hasNext()) {
			Elemento e = elements.next();
			Point p = e.getIntersec(ray);
			if (p != null) {				
				final double squaredDist = Math.abs(p.getZ()
						- ray.getOrig().getZ());// v.getSquaredNorm();
				
				if (squaredDist < minSquaredDist) {
					minSquaredDist = squaredDist;
					closestElement = e;
					intersectionPoint = p;
				}
			}
			elemTemp++;
		}
		// Primeiro raio
		if ((secondaryRay==0)) {
			// Testa se o raio intersecta com a luz antes de qualquer objeto, se
			// esse for o caso, retorna a cor da mesma.
			Iterator<Light> lights = scene.getLightIt();
			while (lights.hasNext()) {
				Light l = lights.next();
				Iterator<Face> faces = l.getGeom().getFacesIt();
				while (faces.hasNext()) {
					Face f = faces.next();
					Point p = f.getInterseccao(ray);
					if (p != null) {
						final double squaredDist = Math.abs(p.getZ()
								- ray.getOrig().getZ());
						if (squaredDist < minSquaredDist) {
							// Caso ele interseccione a luz, retorna a cor da mesma
							Color aquiiiii = l.getColor();
							
							
							
							
							Color a = aquiiiii;
							return new Color(l.getColor());
							//throw new LightInterceptedException(l);
						}
					}

				}

			}

		}
		// Nao interseccionou ninguem
		if (closestElement == null) { 
			if (secondaryRay >0) 
				return new Color(1, 1, 1);
			else
				return new Color(scene.getBack());
		}
		// chama o shade pra esse ponto. Verifica se há sombra e computa o valor de ka+ks+kd
		Color result = this.shadowRay2(ray, closestElement, intersectionPoint);

		if (secondaryRay>2) {
			// Se parar a recursao retorna preto
			//System.out.println(secondaryRay);
			return new Color(1,1,1);
		} else {
			
			Vetor n = closestElement.calcularNormal(intersectionPoint, ray); // A normal ao ponto de interseccao
			
			Color colorSecondaryRay = null;
			// propriedades do material do objeto intersectado
			final Material m = closestElement.getM();
			// Calcula o Ktotal para efetuar o RayTracing
			double kTotal = m.getKs() + m.getKd() + m.getKT();
			double r = random.nextDouble() * kTotal;

			if (r < m.getKd()) { // Raio difuso
				// gera um raio aleatorio
				double phi = Math.acos(Math.sqrt(random.nextDouble()));
				double theta = 2 * Math.PI * random.nextDouble();
				
				Vetor randomVec = n.rotate(intersectionPoint, phi, theta);
				randomVec.normalize();
				if(randomVec.dot(n)<0){
					//randomVec.invertOnOrig();
				}
				//randomVec = n;
				colorSecondaryRay = this.pathTrace(new Ray(
						intersectionPoint, intersectionPoint
								.getAddition(randomVec)), secondaryRay+1);
				colorSecondaryRay.multiplicar(m.getKd());
			} else if (r < m.getKd() + m.getKs()) { // Raio Especular
				Vetor v = new Vetor(intersectionPoint, ray.getOrig());

				Vetor rVec = n.getReflection(v);
				rVec.normalize();

				colorSecondaryRay = this.pathTrace(new Ray(
						intersectionPoint, intersectionPoint
								.getAddition(rVec)), secondaryRay+1);
				colorSecondaryRay.multiplicar(m.getKs());
			} else {
				//Vector3D *V = new Vector3D(ray->direction->x,ray->direction->y,ray->direction->z); 
				Vetor V = new Vetor(intersectionPoint,ray.getOrig()); 
				//V->mult(-1); 
				V.normalize(); 
				
				Ray secondRay = new Ray(ray.getDir(), ray.getOrig());
				
				Vetor auxRay = V.getPerfectReflection(n,V,closestElement.getM().getIr());
				secondRay.setDir(new Point(auxRay.getX(), auxRay.getY(), auxRay.getZ()));  
				if(secondRay.getDir() !=null){
					Vetor aux = new Vetor(0,0,0); 
					aux.setX( secondRay.getDir().getX()); 
					aux.setY(secondRay.getDir().getY()); 
					aux.setZ(secondRay.getDir().getZ()); 
					aux.normalize(); 
					
					Vetor normal2 = n;
					normal2.normalize();
					
					Vetor auxTr = normal2.getPerfectReflection(normal2, aux, closestElement.getM().getIr());
					secondRay.setDir(new Point(auxTr.getX(), auxTr.getY(), auxTr.getZ()));
					colorSecondaryRay = this.pathTrace(secondRay, secondaryRay+1);
					
				}else{ 
					
					Vetor aux = n.getPerfectReflection(n,V);
					secondRay.setDir( new Point(aux.getX(),aux.getY(), aux.getZ()));  
					colorSecondaryRay = this.pathTrace(secondRay,secondaryRay+1); 
					 
				} 
				
				colorSecondaryRay.multiplicar(m.getKT());

			}
			

			result.somar(colorSecondaryRay);
	}
	
	return result;

	}

	
	private Color shadowRay2(final Ray ray, final Elemento element,
			final Point intersectionPoint){
		final Material m = element.getM();
		Vetor normal = element.calcularNormal(intersectionPoint, ray);
		
		double corRambiente = this.scene.getAmbient() * m.getKa() * m.getR();
		double corGambiente = this.scene.getAmbient() * m.getKa() * m.getG();
		double corBambiente = this.scene.getAmbient() * m.getKa() * m.getB();
		
		double corRdifuso = 0.0;
		double corGdifuso = 0.0;
		double corBdifuso = 0.0;
		
		double corEspecular = 0.0;
		
		double R = 0.0;
		double G = 0.0;
		double B = 0.0;
		
		Iterator<Light> lights = scene.getLightIt();
		
		
		//Calculus
		while(lights.hasNext()){
			Light light = lights.next();
			Point pLight = light.getGeom().getAleatoryPointInner();
			Ray l = new Ray(pLight, intersectionPoint);
			
			if(!this.scene.hasShadow(element, l, intersectionPoint)){
				
				
				Vetor lightDir = new Vetor(intersectionPoint, pLight);
				lightDir.normalize();
				
				// Comp difusa
				double cosNL = lightDir.dot(normal);
				//System.out.println(cosNL);
				if(cosNL > 0.0 && cosNL < 1.0){	
					double iDif = /*l.getInt()*/ 1.0* element.getM().getKd() * cosNL;
					corRdifuso += element.getM().getR() * iDif;
					corGdifuso += element.getM().getG() * iDif;
					corBdifuso += element.getM().getB() * iDif;
				}
					
				// Comp especular
				Vetor vetorR = Scene.calcReflection(lightDir, normal, intersectionPoint);
				vetorR.sub(new Vetor(intersectionPoint)); 
				Vetor vetorV = new Vetor(ray.getDir());
				vetorV.sub(new Vetor(ray.getOrig().getX(),ray.getOrig().getY(),ray.getOrig().getZ())); //<=
				vetorV.invertOnOrig();
				vetorV.normalize();
				
				double cosRV = vetorV.dot(vetorR);	
				if(cosRV > 0.0 ){
					double n = Math.pow(cosRV, element.getM().getN());
					double iEspc = (1.0) * (element.getM().getKs()) * n;					
					corEspecular += iEspc;	
				}
			 }		
		}		
		R = corRambiente + corEspecular + corRdifuso;
		G = corGambiente + corEspecular + corGdifuso;
		B = corBambiente + corEspecular + corBdifuso;

		return new Color(R,G,B);
	}

	
	public void desenharImagem() {
		arq.println("P3");
		
		final int width = screen.getWidth();
		final int height = screen.getHeight();

		arq.println(width + " " + height);

		final int factor = 256;
		arq.println(factor);
		for (int i = 0; i < height; i++) {
			for (int j = 0; j < width; j++) {
				arq.println(screen.getColorAt(i, j).toString(factor));
			}
		}
		arq.close();
	} 

	public static void main(String[] args) {
		if (args.length != 0) {
			RayCifePath rayCifePlus = new RayCifePath(args[0]);
			rayCifePlus.executePathTracing();
		} else {
			RayCifePath rayCifePlus = new RayCifePath("cornellroom.sdl");
			rayCifePlus.executePathTracing();
		}

	}
}