package util;

import java.util.ArrayList;
import java.util.List;
import java.util.Random;

import objetosAuxilidares.Ray;


// Face de um objeto geometrico (triangulo)
public class Face {

	private Point v1;
	private Point v2;
	private Point v3;
	private Vetor normal;

	public Face(Point v1, Point v2, Point v3) {
		this.v1 = v1;
		this.v2 = v2;
		this.v3 = v3;
		final Vetor u = new Vetor(this.v2,this.v1);		
		final Vetor v = new Vetor(this.v3,this.v1);
		this.normal =  u.getProdutoVetorial(v);
		normal.normalize();
	}
	
	
	// Encontra a normal usando qualquer ponto
	// pertencente aa face em questao.
	
	public Vetor calcularNormal(final Point p, Ray ray) {
		
		return this.normal;
	}
	
	
	public Vetor calcularNormal2(final Point p, Ray ray) {
		Point p2 = this.v2;
		Point p3 = this.v3;
		
		final Vetor u = new Vetor(p3,p);		
		final Vetor v = new Vetor(p2,p);
		Vetor normal =  u.getProdutoVetorial(v);
		normal.normalize();
		if(normal.dot(new Vetor(p, ray.getDir()))<0.0){
//			System.out.println("Inverteu:");
//			System.out.println(p.getX()+" "+p.getY()+" "+p.getZ());
//			System.out.println(ray.getDir().getX()+" "+ray.getDir().getY()+" "+ray.getDir().getZ());
//			System.out.println("Normal antes:"+normal.getX()+" "+normal.getY()+" "+normal.getZ());
			normal.invertOnOrig();
			//System.out.println("Normal depois:"+normal.getX()+" "+normal.getY()+" "+normal.getZ());
		}	
//		}else{
//			System.out.println("No inverteu:");
//		}
		return normal;
	}

	


	public Point getRandom() {
		Random random = new Random();

		double xPontoLuz = random.nextDouble();
		int yPontoLuzX100 = random.nextInt(100 - ((int) (xPontoLuz * 100)));
		double yPontoLuz = yPontoLuzX100 / 100.0;

		double zPontoLuz = 1 - xPontoLuz - yPontoLuz;

		Point pontoEscolhido = this.getPoint(xPontoLuz, yPontoLuz, zPontoLuz);

		return pontoEscolhido;
	}

	public Point getPoint(double alfa, double beta, double gama) {

		double x = this.v1.getX() * alfa + this.v2.getX() * beta
				+ this.v3.getX() * gama;
		double y = this.v1.getY() * alfa + this.v2.getY() * beta
				+ this.v3.getY() * gama;
		double z = this.v1.getZ() * alfa + this.v2.getZ() * beta
				+ this.v3.getZ() * gama;

		return new Point(x, y, z);
	}

	public Point getInterseccao(final Ray ray) {
		Point origem = ray.getOrig();
		Vetor vetor = new Vetor(ray.getOrig(), ray.getDir());
		vetor.normalize();
		Vetor u = new Vetor(this.v1, this.v2);
		Vetor v = new Vetor(this.v1, this.v3);

		Vetor n = u.getProdutoVetorial(v);

		Vetor dir = vetor;
		Vetor w0 = new Vetor(this.v1, origem);
		double a = -n.dot(w0);
		double b = n.dot(dir);

		if (Math.abs(b) < 0.00000000) {  
			return null; // No existe ponto de interseccao
		}

		
		double r = a / b;
		if (r < 0.0) { 
			return null; //  No existe ponto de interseccao
		}
		dir.svmpy(r);
		Point retorno = origem.getAddition(dir);

		double uu, uv, vv, wu, wv, D;
		uu = u.dot(u);
		uv = u.dot(v);
		vv = v.dot(v);
		Vetor w = new Vetor(this.v1, retorno);
		wu = w.dot(u);
		wv = w.dot(v);
		D = uv * uv - uu * vv;

		double s, t;
		s = (uv * wv - vv * wu) / D;
		if (s < 0.0 || s > 1.0) 
			return null; //  No existe ponto de interseccao
		t = (uv * wu - uu * wv) / D;
		if (t < 0.0 || (s + t) > 1.0) 
			return null; //  No existe ponto de interseccao

		return retorno; // Existe ponto de interseccao
	}
	

	public Point getV1() {
		return v1;
	}

	public Point getV2() {
		return v2;
	}

	public Point getV3() {
		return v3;
	}

	public Point getBaricentro() {
		return new Point((v1.getX() + v2.getX() + v3.getX()) / 3.0, (v1.getY()
				+ v2.getY() + v3.getY()) / 3.0, (v1.getZ() + v2.getZ() + v3
				.getZ()) / 3.0);
	}

	public List<Point> getPoints() {
		List<Point> result = new ArrayList<Point>();
		result.add(this.getBaricentro());
		result.add(v1);
		result.add(v2);
		result.add(v3);
		return result;
	}
	
}