package util;



import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;


import objetosAuxilidares.Elemento;
import objetosAuxilidares.Ray;


// Representacao dos aspectos da cena
public class Scene {

	private Color background; // Cor de fundo
	private double ambient; // Intensidade de luz ambiente
	private Point eye; // Camera
	private List<List<Elemento>> objs; // Objetos da cena
	private List<Light> lights; // Luzes
		
	
	public Scene() {	
		this.background = null;
		this.ambient = 0;
		this.eye = null;
		this.objs = new ArrayList<List <Elemento>>();
		this.lights = new ArrayList<Light>();	
	}

public static Vetor calcReflection(Vetor rin, Vetor normal,  Point intersect){
		
		normal.normalize();
		rin.normalize();
		
		double cos = normal.dot(rin);
		if(cos < 0) cos = 0;
		Vetor rout = new Vetor(normal);
		rout.svmpy(2);
		rout.svmpy(cos);
		rout.sub(rin);
		rout.normalize();
		rout.add(new Vetor(intersect)); 
		return rout;
		
	}
	
	
	public Color getBack() {
		return this.background;
	}

	public void setBack(Color arg0) {
		this.background = arg0;
	}

	public double getAmbient() {
		return this.ambient;
	}

	public void setAmbient(double arg0) {
		this.ambient = arg0;
	}

	public Point getEye() {
		return this.eye;
	}

	public void setEye(Point eye) {
		this.eye = eye;
	}

	public Elemento getObjectAt(int arg0) {
		List<Elemento> elms = new ArrayList<Elemento>();
		for(List<Elemento> objs : this.objs){
			for(Elemento itQuad: objs){
				elms.add(itQuad);
			}
		}
		return elms.get(arg0);
	}

	public void addObject(List<Elemento> arg0) {
		this.objs.add(arg0);
	}

	
	public Iterator<Elemento> getObjIt() {
		List<Elemento> elms = new ArrayList<Elemento>();
		for(List<Elemento> objs : this.objs){
			for(Elemento itQuad: objs){
				elms.add(itQuad);
			}
		}
		return elms.iterator();
	}

	public Light getLightAt(int arg0) {
		return this.lights.get(arg0);
	}

	public void addLight(Light arg0) {
		this.lights.add(arg0);
	}

	
	public Iterator<Light> getLightIt() {
		return this.lights.iterator();
	}
	
	
	
	
	public boolean hasShadow(Elemento quad, Ray l, Point intersec){
		for(List<Elemento> obj : this.objs){
			for(Elemento itQuad: obj){
				if(!obj.contains(quad)){
					
					Point inters2 = itQuad.getIntersec(l);
					double dist = -1;
					if(inters2 != null && itQuad!= quad ){
						dist = (new Vetor(l.getOrig(),inters2)).getSquaredNorm();
						
					}
					
					
					if((dist > 0.001) && itQuad!= quad && obj.size()!=2){
						return true;
					}
				}
			}
		}
		
		return false;
	}
	
	
	public Retorno hasObjetoIntersetador(final Ray r) {
		Retorno ret = null;
		List<List <Elemento>> objetos = this.objs;
		
		for(List<Elemento> objs : objetos){
			for (Elemento objeto : objs) {
				Point intersecao = objeto.getIntersec(r);
				if (intersecao != null) {
					ret = new Retorno();
					ret.objetoIntersectado = objeto;
					ret.pontoIntersecao = intersecao;
					//return ret;
				}
			}
		}

		return ret;
	}
	
	public class Retorno {
		public Point pontoIntersecao;
		public Elemento objetoIntersectado;
	}	
	
	
	public boolean interceptsAtLeastOneElement(final Ray ray, final Elemento exception){					
		for(List <Elemento> objetos: this.objs){
			for(Elemento e: objetos) {	
					Point p = e.getIntersec(ray);
					if(e == exception){
						if(p != null){
							Vetor v = new Vetor(p,ray.getOrig());
							if(v.getSquaredNorm() > 0.0000001d){
								return true;
							}
						}				
					}else{
						if (p != null) {
							return true;
						}else{
						
						}
					}
			}
		}
		return false;		
	}

}