package util;

public class Vetor {

	
	private double x;
	private double y;
	private double z;
	
	
	public Vetor(double x, double y, double z) {
		this.x = x;
		this.y = y;
		this.z = z;
	}
	
	public Vetor(Point p) {
		this.x = p.getX();
		this.y = p.getY();
		this.z = p.getZ();
	}
	
	public Vetor(Point a, Point b) {
		this.x = b.getX() - a.getX();
		this.y = b.getY() - a.getY();
		this.z = b.getZ() - a.getZ();
	}
	
	//Deep copy constructor
	public Vetor(Vetor v) {
		this.x = v.x;
		this.y = v.y;
		this.z = v.z;
	}
	
	//Getters and setters
	public double getX() {
		return this.x;
	}
	public void setX(double x) {
		this.x = x;
	}
	public double getY() {
		return this.y;
	}
	public void setY(double y) {
		this.y = y;
	}
	public double getZ() {
		return this.z;
	}
	public void setZ(double z) {
		this.z = z;
	}
	
	public void set(Point a, Point b){
		this.x = b.getX() - a.getX();
		this.y = b.getY() - a.getY();
		this.z = b.getZ() - a.getZ();
	}
	
	//This routine normalizes the vector
//	public void normalize() {
//		
//		// Temporary denominator
//		double denom = 0;
//		
//		//Absolute value of vector's coordinates
//		double x = (this.x > 0.0) ? this.x : - this.x;
//		double y = (this.y > 0.0) ? this.y : - this.y;
//		double z = (this.z > 0.0) ? this.z : - this.z;
//		
//		//(x,y,z) / sqrt(x*x + y*y + z*z)		
//		if(x > y){
//			if(x > z){
//				y = y / x;
//				z = z / x;
//				denom = 1.0 / (x * Math.sqrt(1.0 + y * y + z * z));
//			}else{				//z > x > y
//				if(1.0 + z > 1.0){
//					y = y / z;
//					x = x / z;
//					denom = 1.0 / (z * Math.sqrt(1.0 + y * y + x * x));
//				}
//			}
//		}else{
//			if(y > z){
//				z = z / y;
//				x = x / y;
//				denom = 1.0 / (y * Math.sqrt(1.0 + z * z + x * x));
//			}else{				//x < y < z
//				if(1.0 + z > 1.0){
//					y = y / z;
//					x = x / z;
//					denom = 1.0 / (z * Math.sqrt(1.0 + y * y + x * x));
//				}
//			}
//		}
//		
//		if(1.0 + x + y + z > 1.0){
//			this.svmpy(denom);
//		}
//	}							//End procedure normalize
//	
	
	//This routine normalizes the vector
	public void normalize() {
		double n = Math.sqrt(x*x + y*y + z*z);
		x /= n;
		y /= n;
		z /= n;		
	}							//End procedure normalize
	
	public double getSquaredNorm(){
		return x*x + y*y + z*z;
	}
	
	public Vetor getProdutoVetorial(final Vetor v){
		//(v1.y * v2.z) - (v1.z * v2.y)
		//(v1.z * v2.x) - (v1.x * v2.z)
		//(v1.x * v2.y) - (v1.y * v2.x)
		
		final double x2 = v.getX();
		final double y2 = v.getY();
		final double z2 = v.getZ();		
		return new Vetor(y * z2 - z * y2, z * x2 - x * z2, x * y2 - y * x2);
		//return new Vetor(y * z2 - z * y2, x * z2 - z * x2, x * y2 - y * x2);
	}
	
	//This routine multiples the vector by a constant
	//(Produto escalar)
	public void svmpy(double scalar) {
		this.x *= scalar;
		this.y *= scalar;
		this.z *= scalar;
	}							//End procedure svmpy
	
	//This routine computes the dot product between this vector and the
	// parameter
	//(Produto Interno)
	public double dot(Vetor v) {
		return this.x * v.x + this.y * v.y + this.z * v.z;
	}
	
	//This routine adds two vectors
	public Vetor vadd(Vetor a, Vetor b) {
		
		//Vector structure to hold result
		Vetor result = new Vetor(a);
		result.sub(b);
		
		return result;
	}							//End procedure vadd
	
	//Essa rotina adiciona esse vetor ao vetor par�metro
	//(adicional)
	public void add(Vetor v) {
		this.x += v.x;
		this.y += v.y;
		this.z += v.z;
	}
	
	//This routine subtracts two vectors
	public Vetor vsub(Vetor a, Vetor b) {
		
		//Vector structure to hold result
		Vetor result = new Vetor(a);
		result.sub(b);	
		
		return result;
	}							//End procedure vsub
	
	//Essa rotina subtrai desse vetor o vetor par�metro
	//(adicional)
	public void sub(Vetor v) {
		this.x -= v.x;
		this.y -= v.y;
		this.z -= v.z;
	}
	
	//Essa rotina inverte o valor das coordenadas do vetor
	public void invertOnOrig() {
		this.x = -this.x;
		this.y = -this.y;
		this.z = -this.z;
	}
	
	public Vetor getReflection(Vetor v){		
		Vetor result = new Vetor(this);
		final double factor = 2*this.dot(v);
		result.svmpy(factor);
		result.sub(v);		
//		if(result.dot(v)<0.0){
//			result.invertOnOrig();
//		}
		return result;
	}
	
	public Vetor getPerfectReflection(Vetor normal, Vetor vector){
		Vetor response;
		double factor = 2*(normal.dot(vector));
		response = new Vetor(normal.getX(),normal.getY(),normal.getZ());
		response.svmpy(factor);
		response.sub(vector);
		return response;
	}
	
	public Vetor getPerfectReflection(Vetor normal, Vetor direction, double ir){
		Vetor response = null;
		double temp;
		Vetor newDir = new Vetor(direction.getX(), direction.getY(), direction.getZ()); 
		newDir.svmpy(-1);

		double cosNormalDirection = 1.0/newDir.dot(normal);

		newDir.svmpy(-1);
		newDir.svmpy(cosNormalDirection);

		temp = (ir * ir) * (newDir.dot(newDir));
		newDir.add(normal);
		temp = temp - (newDir.dot(newDir));
		if(temp >= 0){
			double alpha = 1.0/Math.sqrt(temp);
			newDir.svmpy(alpha);
			newDir.sub(normal);
			response = new Vetor(newDir.x,newDir.y,newDir.z);
		}

		return response;
	}
	
	public String toString() {
		return "[" + this.x + " "
				+ this.y + " " + this.z + "]";
	}	
	
	
	public Vetor rotate(Vetor normal, double phi, double teta){
		Vetor response ;//= new Vector3D();
		double newZ;
		newZ = (-Math.pow(normal.getX(), 2) - Math.pow(normal.getY(), 2)) / normal.getZ();
		
		Vetor vec1 = new Vetor(normal.getX(),normal.getZ(),newZ);
		vec1.normalize();

		double xRotation = Math.cos(phi) * Math.cos(teta);
		double yRotation = -Math.sin(teta) * Math.cos(phi);
		double zRotation = -Math.sin(phi);
		double responseX, responseY, responseZ;

		Vetor vector2 = normal.getProdutoVetorial(vec1);
		vector2.normalize();
		
		responseX = (xRotation * vector2.getX()) + (yRotation * vec1.getX()) + (zRotation * normal.getX());
		responseY = (xRotation * vector2.getY()) + (yRotation * vec1.getY()) + (zRotation * normal.getY());
		responseZ = (xRotation * vector2.getZ()) + (yRotation * vec1.getZ()) + (zRotation * normal.getZ());
		
		
		response = new Vetor(responseX, responseY, responseZ);
		return response;
	}
	
	public Vetor rotate(Point origin, double phi, double teta){
		Vetor response = new Vetor(0,0,0);
		response.setX( origin.getX()); // nao muda pela rotacao
	     response.setY(origin.getY() * Math.cos(phi) - origin.getZ() * Math.sin(phi));
	     response.setZ(origin.getZ() * Math.sin(phi) + origin.getZ() * Math.cos(phi));
	     //girando em torno do outro eixo
	     response.setX(response.getX() * Math.cos(teta) + response.getZ() * Math.sin(teta));
	     //ret.Y permanece o mesmo.
	     response.setZ( (-response.getX()) * Math.sin(teta) + response.getZ() * Math.cos(teta));
	     return response;
//		Vector3D *response ;//= new Vector3D();
//		double newZ;
//		newZ = (-powf(normal->x, 2) - powf(normal->y, 2)) / normal->z;
//		
//		Vector3D *vector1 = new Vector3D(normal->x,normal->z,newZ);
//		vector1->normalize();
//
//		double xRotation = cos(phi) * cos(theta);
//		double yRotation = -sin(theta) * cos(phi);
//		double zRotation = -sin(phi);
//		double responseX, responseY, responseZ;
//
//		Vector3D *vector2 = normal->cross(vector1);
//		vector2->normalize();
//		
//		responseX = (xRotation * vector2->x) + (yRotation * vector1->x) + (zRotation * normal->x);
//		responseY = (xRotation * vector2->y) + (yRotation * vector1->y) + (zRotation * normal->y);
//		responseZ = (xRotation * vector2->z) + (yRotation * vector1->z) + (zRotation * normal->z);
//		
//		delete vector1; delete vector2;
//
//		response = new Vector3D(responseX, responseY, responseZ);
//		return response;
	}
}