/*
 * Created on 05/03/2008
 */
package numerics;

import java.util.ArrayList;
import java.util.List;
import java.util.Random;

import objectDefinitions.Ray;

/**
 * @author Marlus Barbosa de Souza - mbs2
 * @author Artur Lira dos Santos - als3
 */
public class Face {

	private Point v1;
	private Point v2;
	private Point v3;
	private Vec normal;

	public Face(Point v1, Point v2, Point v3) {
		this.v1 = v1;
		this.v2 = v2;
		this.v3 = v3;
		final Vec u = new Vec(this.v2,this.v1);		
		final Vec v = new Vec(this.v3,this.v1);
		this.normal =  u.getCrossProduct(v);
		normal.normalize2();
	}
	
	public Vec findNormal(final Point p) {		
		return new Vec(normal);
	}

	


	public Point getRandom() {
		Random random = new Random();

		double xPontoLuz = random.nextDouble();
		int yPontoLuzX100 = random.nextInt(100 - ((int) (xPontoLuz * 100)));
		double yPontoLuz = yPontoLuzX100 / 100.0;

		double zPontoLuz = 1 - xPontoLuz - yPontoLuz;

		Point pontoEscolhido = this.getPoint(xPontoLuz, yPontoLuz, zPontoLuz);

		return pontoEscolhido;
	}

	public Point getPoint(double alfa, double beta, double gama) {

		double x = this.v1.getX() * alfa + this.v2.getX() * beta
				+ this.v3.getX() * gama;
		double y = this.v1.getY() * alfa + this.v2.getY() * beta
				+ this.v3.getY() * gama;
		double z = this.v1.getZ() * alfa + this.v2.getZ() * beta
				+ this.v3.getZ() * gama;

		return new Point(x, y, z);
	}

	public Point getIntersectionPoint(final Ray ray) {
		//System.out.println(ray.getOrig()+ " dir:" +ray.getDir());
		//System.out.println(this.getV1() + " y" + this.getV2()+ " z" + this.getV3());
		Point origem = ray.getOrig();
		Vec vetor = new Vec(ray.getOrig(), ray.getDir());
		vetor.normalize2(); // vetor.normalize2();

		// get triangle edge vectors and plane normal
		Vec u = new Vec(this.v1, this.v2);
		Vec v = new Vec(this.v1, this.v3);

		Vec n = u.getCrossProduct(v);

		Vec dir = vetor; // ray direction vector
		Vec w0 = new Vec(this.v1, origem);
		double a = -n.dot(w0);
		double b = n.dot(dir);

		if (Math.abs(b) < 0.00000000) { // ray is parallel to triangle plane 1e-10
			if (a == 0) {
				return null; // ray lies in triangle plane
			} else {
				return null; // ray disjoint from plane
			}
		}

		// get intersect point of ray with triangle plane
		double r = a / b;
		if (r < 0.0) { // ray goes away from triangle
			return null; // => no intersect
		}
		// for a segment, also test if (r > 1.0) => no intersect

		dir.svmpy(r);
		Point retorno = origem.getAddition(dir); // intersect point of ray
													// and plane

		// is I inside T?
		double uu, uv, vv, wu, wv, D;
		uu = u.dot(u);
		uv = u.dot(v);
		vv = v.dot(v);
		Vec w = new Vec(this.v1, retorno);
		wu = w.dot(u);
		wv = w.dot(v);
		D = uv * uv - uu * vv;

		// get and test parametric coords
		double s, t;
		s = (uv * wv - vv * wu) / D;
		if (s < 0.0 || s > 1.0) // I is outside T
			return null;
		t = (uv * wu - uu * wv) / D;
		if (t < 0.0 || (s + t) > 1.0) // I is outside T
			return null;

		return retorno; // I is in T
	}
	
	// retorna null se n�o intersectar.
	public Point getIntersectionPoint2(final Ray r) {	 

		final Vec d = new Vec(r.getOrig(), r.getDir());
		final double q = d.dot(normal);
		final Vec po = new Vec(this.v1, r.getOrig());
		final double t = -(po.dot(normal) / q);
		d.svmpy(t);
		Point p = new Point(r.getOrig());
		p.add(d);
		if (this.intersects(p))
			return p;
		else {
			return null;
		}
	}

	// checa corretamente apenas no caso do ponto p pertencer ao plano da face	
	private boolean intersects(final Point p) {
		final Vec v = new Vec(v1, v2);
		final Vec u = new Vec(v1, p);
		if (v.getCrossProduct(u).dot(normal) < 0)
			return false;
		else {
			v.set(v2, v3);
			u.set(v2, p);
			if (v.getCrossProduct(u).dot(normal) < 0)
				return false;
			else {
				v.set(v3, v1);
				u.set(v3, p);
				return v.getCrossProduct(u).dot(normal) >= 0;
			}
		}
	}

	public Point getV1() {
		return v1;
	}

	public Point getV2() {
		return v2;
	}

	public Point getV3() {
		return v3;
	}

	public Point getBarycenter() {
		return new Point((v1.getX() + v2.getX() + v3.getX()) / 3.d, (v1.getY()
				+ v2.getY() + v3.getY()) / 3.d, (v1.getZ() + v2.getZ() + v3
				.getZ()) / 3.d);
	}

	public List<Point> getPoints() {
		List<Point> result = new ArrayList<Point>();
		result.add(this.getBarycenter());
		result.add(v1);
		result.add(v2);
		result.add(v3);
		return result;
	}
	

	public static void main(String... args) {
		Point p0 = new Point(0, 0, 0);
		Point p1 = new Point(10, 0, 0);
		Point p2 = new Point(0, 10, 0);
		Face f = new Face(p0, p1, p2);
		Point orig = new Point(5, 5, 10);
		Point dir = new Point(5, 5, -10);
		Point p = f.getIntersectionPoint(new Ray(dir,orig));
		System.out.println(p);
		System.out.println(f.findNormal(p));
		
		//(-0.16947362697153953,0.16947362697153953,4.729146056544949)
		//(0.0,0.0,5.7)



		//3.8220 3.8416 -32.7600 3.8220 3.8416 -16.5900 -3.8220 3.8416 -16.5900
		
		p0 = new Point(3.8220, 3.8416, -32.7600);
		p1 = new Point(3.8220, 3.8416, -16.5900);
		p2 = new Point(-3.8220, 3.8416, -16.5900);
		Point p3 = new Point(-3.8220, 3.8416, -32.7600);
		f = new Face(p0,p1,p2);
		Face f2 = new Face(p0,p2,p3);
		
		
		orig = new Point(0.0,0.0,5.7);
		
		dir = new Point(-0.16947362697153953,0.16947362697153953,4.729146056544949);
		p = f.getIntersectionPoint(new Ray(dir, orig));
		System.out.println(p);
		System.out.println(f.findNormal(p));
		
		
		p = f2.getIntersectionPoint(new Ray(dir, orig));
		System.out.println(p);
		System.out.println(f.findNormal(p));
	
	}
}

/**
// Copyright 2001, softSurfer (www.softsurfer.com)
// This code may be freely used and modified for any purpose
// providing that this copyright notice is included with it.
// SoftSurfer makes no warranty for this code, and cannot be held
// liable for any real or imagined damage resulting from its use.
// Users of this code must verify correctness for their application.

// Assume that classes are already given for the objects:
//    Point and Vector with
//        coordinates {float x, y, z;}
//        operators for:
//            == to test equality
//            != to test inequality
//            (Vector)0 = (0,0,0)         (null vector)
//            Point  = Point ± Vector
//            Vector = Point - Point
//            Vector = Scalar * Vector    (scalar product)
//            Vector = Vector * Vector    (cross product)
//    Line and Ray and Segment with defining points {Point P0, P1;}
//        (a Line is infinite, Rays and Segments start at P0)
//        (a Ray extends beyond P1, but a Segment ends at P1)
//    Plane with a point and a normal {Point V0; Vector n;}
//    Triangle with defining vertices {Point V0, V1, V2;}
//    Polyline and Polygon with n vertices {int n; Point *V;}
//        (a Polygon has V[n]=V[0])
//===================================================================

#define SMALL_NUM  0.00000001 // anything that avoids division overflow
// dot product (3D) which allows vector operations in arguments
#define dot(u,v)   ((u).x * (v).x + (u).y * (v).y + (u).z * (v).z)

// intersect_RayTriangle(): intersect a ray with a 3D triangle
//    Input:  a ray R, and a triangle T
//    Output: *I = intersection point (when it exists)
//    Return: -1 = triangle is degenerate (a segment or point)
//             0 = disjoint (no intersect)
//             1 = intersect in unique point I1
//             2 = are in the same plane
int
intersect_RayTriangle( Ray R, Triangle T, Point* I )
{
    Vector    u, v, n;             // triangle vectors
    Vector    dir, w0, w;          // ray vectors
    float     r, a, b;             // params to calc ray-plane intersect

    // get triangle edge vectors and plane normal
    u = T.V1 - T.V0;
    v = T.V2 - T.V0;
    n = u * v;             // cross product
    if (n == (Vector)0)            // triangle is degenerate
        return -1;                 // do not deal with this case

    dir = R.P1 - R.P0;             // ray direction vector
    w0 = R.P0 - T.V0;
    a = -dot(n,w0);
    b = dot(n,dir);
    if (fabs(b) < SMALL_NUM) {     // ray is parallel to triangle plane
        if (a == 0)                // ray lies in triangle plane
            return 2;
        else return 0;             // ray disjoint from plane
    }

    // get intersect point of ray with triangle plane
    r = a / b;
    if (r < 0.0)                   // ray goes away from triangle
        return 0;                  // => no intersect
    // for a segment, also test if (r > 1.0) => no intersect

    *I = R.P0 + r * dir;           // intersect point of ray and plane

    // is I inside T?
    float    uu, uv, vv, wu, wv, D;
    uu = dot(u,u);
    uv = dot(u,v);
    vv = dot(v,v);
    w = *I - T.V0;
    wu = dot(w,u);
    wv = dot(w,v);
    D = uv * uv - uu * vv;

    // get and test parametric coords
    float s, t;
    s = (uv * wv - vv * wu) / D;
    if (s < 0.0 || s > 1.0)        // I is outside T
        return 0;
    t = (uv * wu - uu * wv) / D;
    if (t < 0.0 || (s + t) > 1.0)  // I is outside T
        return 0;

    return 1;                      // I is in T
}

**/