/*
 * Created on 05/03/2008
 */
package numerics;




/**
 * @author Marlus Barbosa de Souza - mbs2
 * @author Artur Lira dos Santos - als3
 */

//From quadric.h -> Prototypes for quadric.cpp

//Vector structure
public class Vec {

	//Coefficients of vector direction
	private double x;
	private double y;
	private double z;
	
	//Vector constructor
	public Vec(double x, double y, double z) {
		this.x = x;
		this.y = y;
		this.z = z;
	}
	
	public Vec(Point p) {
		this.x = p.getX();
		this.y = p.getY();
		this.z = p.getZ();
	}
	
	public Vec(Point a, Point b) {
		this.x = b.getX() - a.getX();
		this.y = b.getY() - a.getY();
		this.z = b.getZ() - a.getZ();
	}
	
	//Deep copy constructor
	public Vec(Vec v) {
		this.x = v.x;
		this.y = v.y;
		this.z = v.z;
	}
	
	//Getters and setters
	public double getX() {
		return this.x;
	}
	public void setX(double x) {
		this.x = x;
	}
	public double getY() {
		return this.y;
	}
	public void setY(double y) {
		this.y = y;
	}
	public double getZ() {
		return this.z;
	}
	public void setZ(double z) {
		this.z = z;
	}
	
	public void set(Point a, Point b){
		this.x = b.getX() - a.getX();
		this.y = b.getY() - a.getY();
		this.z = b.getZ() - a.getZ();
	}
	
	//This routine normalizes the vector
	public void normalize() {
		
		// Temporary denominator
		double denom = 0;
		
		//Absolute value of vector's coordinates
		double x = (this.x > 0.0) ? this.x : - this.x;
		double y = (this.y > 0.0) ? this.y : - this.y;
		double z = (this.z > 0.0) ? this.z : - this.z;
		
		//(x,y,z) / sqrt(x*x + y*y + z*z)		
		if(x > y){
			if(x > z){
				y = y / x;
				z = z / x;
				denom = 1.0 / (x * Math.sqrt(1.0 + y * y + z * z));
			}else{				//z > x > y
				if(1.0 + z > 1.0){
					y = y / z;
					x = x / z;
					denom = 1.0 / (z * Math.sqrt(1.0 + y * y + x * x));
				}
			}
		}else{
			if(y > z){
				z = z / y;
				x = x / y;
				denom = 1.0 / (y * Math.sqrt(1.0 + z * z + x * x));
			}else{				//x < y < z
				if(1.0 + z > 1.0){
					y = y / z;
					x = x / z;
					denom = 1.0 / (z * Math.sqrt(1.0 + y * y + x * x));
				}
			}
		}
		
		if(1.0 + x + y + z > 1.0){
			this.svmpy(denom);
		}
	}							//End procedure normalize
	
	
	//This routine normalizes the vector
	public void normalize2() {
		double n = Math.sqrt(x*x + y*y + z*z);
		x /= n;
		y /= n;
		z /= n;		
	}							//End procedure normalize
	
	public double getSquaredNorm(){
		return x*x + y*y + z*z;
	}
	
	public Vec getCrossProduct(final Vec v){
		final double x2 = v.getX();
		final double y2 = v.getY();
		final double z2 = v.getZ();		
		return new Vec(y * z2 - z * y2, x * z2 - z * x2, x * y2 - y * x2);
	}
	
	//This routine multiples the vector by a constant
	//(Produto escalar)
	public void svmpy(double scalar) {
		this.x *= scalar;
		this.y *= scalar;
		this.z *= scalar;
	}							//End procedure svmpy
	
	//This routine computes the dot product between this vector and the
	// parameter
	//(Produto Interno)
	public double dot(Vec v) {
		return this.x * v.x + this.y * v.y + this.z * v.z;
	}
	
	//This routine adds two vectors
	public Vec vadd(Vec a, Vec b) {
		
		//Vector structure to hold result
		Vec result = new Vec(a);
		result.sub(b);
		
		return result;
	}							//End procedure vadd
	
	//Essa rotina adiciona esse vetor ao vetor parmetro
	//(adicional)
	public void add(Vec v) {
		this.x += v.x;
		this.y += v.y;
		this.z += v.z;
	}
	
	//This routine subtracts two vectors
	public Vec vsub(Vec a, Vec b) {
		
		//Vector structure to hold result
		Vec result = new Vec(a);
		result.sub(b);	
		
		return result;
	}							//End procedure vsub
	
	//Essa rotina subtrai desse vetor o vetor parmetro
	//(adicional)
	public void sub(Vec v) {
		this.x -= v.x;
		this.y -= v.y;
		this.z -= v.z;
	}
	
	//Essa rotina inverte o valor das coordenadas do vetor
	//(em relao  origem)
	public void invertOnOrig() {
		this.x = -this.x;
		this.y = -this.y;
		this.z = -this.z;
	}
	
	public Vec getReflection(Vec v){		
		Vec result = new Vec(this);
		final double factor = 2*this.dot(v);
		result.svmpy(factor);
		result.sub(v);		
		return result;
	}
	
	public String toString() {
		return "[" + this.x + " "
				+ this.y + " " + this.z + "]";
	}	
}