package objectDefinitions.Element;

import numerics.Point;
import numerics.Quadric;
import numerics.Vec;
import objectDefinitions.Mat;
import objectDefinitions.QuadCoef;
import objectDefinitions.Ray;

public class QuadricElement extends Element {
	private QuadCoef coef;			//Gives the coefficients of the quadric	
	
	public QuadricElement(QuadCoef coef,Mat m) {
		super(m);
		this.coef = coef;
	}
	
	public QuadCoef getCoef() {
		return coef;
	}

	@Override
	public boolean isGeometric() {
		return false;
	}
	
	@Override
	public boolean isQuadric() {
		return true;
	}

	@Override
	public Point getIntersection(Ray ray) {
		final double dist = Quadric.intersect(ray,coef);
		if(dist <= 0)
			return null;
		Vec v = new Vec(ray.getOrig(),ray.getDir());
		v.normalize2();
		v.svmpy(dist);
		Point result = new Point(ray.getOrig());
		result.add(v);
		return result;
	}

	@Override
	public Vec findNormal(final Point p) {		
		Vec result = coef.getNormal(p);
		result.normalize2();
		return result;
	}
	
}
