/*
 * Created on 05/03/2008
 */
package objectDefinitions;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Random;

import numerics.Face;
import numerics.Point;
import numerics.Vec;
/**
 * @author Marlus Barbosa de Souza - mbs2
 * @author Artur Lira dos Santos - als3
 */
public class ObjectGeometry {
	
	private List<Point> points;	//Points of the object geometry
	private List<Face> faces;		//Indexes of the points that compose the faces in the object geometry
	
	//Constructor
	public ObjectGeometry(List<Point> points, List<Face> faces){
		this.points = points;
		this.faces = faces;		
	}
	
	//Returns the point at the given index
	public Point getPointAt(int arg0){
		return this.points.get(arg0);
	}
	//Adds the point at the end of the vector
	public void addPoint(Point arg0){
		this.points.add(arg0);
	}
	//Returns the face at the given index
	public Face getFaceAt(int arg0){
		return this.faces.get(arg0);
	}
	//Adds the face at the end of the vector
	public void addFace(Face arg0){
		this.faces.add(arg0);
	}
	
	public Vec findNormal(final Point p) {
		for(Face face : faces){
			return face.findNormal(p);
		}
		return null;
	}
	
	public List<Point> getBoundingPlusBarycenterPoints(){
		
		List<Point> result = new ArrayList<Point>();
		/*Random r = new Random();
		for(int i = 0; i < 100; i++){			
			int index = r.nextInt(faces.size());
			Face f = faces.get(index);
			result.add(f.getRandom());
		}*/
		
		
		
		/*
		 * Point pResult = new Point(0,0,0);
		 * for(Point p : result){
			pResult.setX(pResult.getX() + p.getX());
			pResult.setY(pResult.getY() + p.getY());
			pResult.setZ(pResult.getZ() + p.getZ());
		}
		pResult.setX(pResult.getX() / result.size());
		pResult.setY(pResult.getY() / result.size());
		pResult.setZ(pResult.getZ() / result.size());*/
		
		result.addAll(points);
		for(Face face : faces){
			result.add(face.getBarycenter());			
		}

		
		
		
		
		return result;
	}
	
	public Iterator<Face> getFacesIt(){
		return this.faces.iterator();
	}
	
	public Point getIntersection(final Ray ray){		
		for(Face face : faces){
			final Point p = face.getIntersectionPoint(ray);
			if(p != null)
				return p;
		}
		return null;		
	}

	public Face getFace(final Ray ray) {
		for(Face face : faces){			
			if(face.getIntersectionPoint(ray) != null){
				return face;
			}
		}
		return null;
	}
	
	public List<ObjectGeometry> getObjects(){
		List<ObjectGeometry> result = new ArrayList<ObjectGeometry>();
		for(Face f: faces){
			List<Point> points = new ArrayList<Point>(this.points);	
			List<Face> faces = new ArrayList<Face>();
			faces.add(f);
			result.add(new ObjectGeometry(points,faces));
		}		
		return result;
	}
	public Point getAleatoryPointInner(){
		Point auxReturn = null;
		Random rd = new Random();
		int numFaces = this.faces.size();
		Face face = this.faces.get(rd.nextInt(numFaces));
		Point v1 = face.getV1();
		Point v2 = face.getV2();
		Point v3 = face.getV3();
		
		Vec vec1 = new Vec(v1, v2);
		Vec vec2 = new Vec(v2, v3);
		double scalar = rd.nextDouble();
		vec2.svmpy(scalar);
		vec1.add(vec2);
		scalar = rd.nextDouble();
		vec1.svmpy(scalar);
		vec1.add(new Vec(v1));
		
		auxReturn =new Point(vec1.getX(), vec1.getY(), vec1.getZ());
		
		return auxReturn;
	}
}