/*
 * Created on 05/03/2008
 */
package sceneDefinitions;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import numerics.Point;
import numerics.Vec;
import objectDefinitions.Ray;
import objectDefinitions.Element.Element;

/**
 * @author Marlus Barbosa de Souza - mbs2
 * @author Artur Lira dos Santos - als3
 */

// Classe que representa os aspectos da cena
public class Scene {

	private Color background; // Cor de fundo
	private double ambient; // Intensidade de luz ambiente
	private Point eye; // Posio do olho do observador (cmera)
	private List<Element> objs; // Vetor que armazena os objetos da cena
	private List<Light> lights; // Vetor que armazena as luzes
		
	// Construtor da classe Cena, que inicializa alguns atributos
	public Scene() {	
		this.background = null;
		this.ambient = 0;
		this.eye = null;
		this.objs = new ArrayList<Element>();
		this.lights = new ArrayList<Light>();	
	}

	// Getters and setters
	public Color getBack() {
		return this.background;
	}

	public void setBack(Color arg0) {
		this.background = arg0;
	}

	public double getAmbient() {
		return this.ambient;
	}

	public void setAmbient(double arg0) {
		this.ambient = arg0;
	}

	public Point getEye() {
		return this.eye;
	}

	public void setEye(Point eye) {
		this.eye = eye;
	}

	public Element getObjectAt(int arg0) {
		return this.objs.get(arg0);
	}

	public void addObject(Element arg0) {
		this.objs.add(arg0);
	}

	// Retorna um Iterator para os objetos
	public Iterator<Element> getObjIt() {
		return this.objs.iterator();
	}

	public Light getLightAt(int arg0) {
		return this.lights.get(arg0);
	}

	public void addLight(Light arg0) {
		this.lights.add(arg0);
	}

	// Retorna um Iterator para as luzes
	public Iterator<Light> getLightIt() {
		return this.lights.iterator();
	}
	
	public Retorno hasObjetoIntersetador(final Ray r) {
		Retorno ret = null;
		List<Element> objetos = this.objs;

		for (Element objeto : objetos) {
			Point intersecao = objeto.getIntersection(r);
			if (intersecao != null) {
				ret = new Retorno();
				ret.objetoIntersectado = objeto;
				ret.pontoIntersecao = intersecao;
			}
		}

		return ret;
	}
	
	public class Retorno {
		public Point pontoIntersecao;
		public Element objetoIntersectado;
	}	
	
	
	public boolean interceptsAtLeastOneElement(final Ray ray, final Element exception){					
		for(Element e: this.objs) {	
			//if(e.getM() != exception.getM()){
				Point p = e.getIntersection(ray);
				if(e == exception){
					if(p != null){
						Vec v = new Vec(p,ray.getOrig());
						if(v.getSquaredNorm() > 0.0000001d){
							return true;
						}
					}				
				}else{
					if (p != null) {
						return true;
					}else{
					
					}
				}
			//}		
			
		}		
		return false;		
	}

}